use std::{
    error::Error,
    fs::{read, write},
    path::PathBuf,
    process::exit,
};

use clap::Parser;

use radicle::{git::RefString, prelude::RepoId};
use radicle_git_ext::Oid;

use radicle_ci_broker::{
    db2::{Db, QueueId},
    error::BrokerError,
    event::BrokerEvent,
};

fn main() {
    if let Err(e) = fallible_main() {
        eprintln!("ERROR: {}", e);
        let mut e = e.source();
        while let Some(source) = e {
            eprintln!("caused by: {}", source);
            e = source.source();
        }
        exit(1);
    }
}

fn fallible_main() -> Result<(), BrokerError> {
    pretty_env_logger::init();

    let args = Args::parse();
    args.run()?;

    Ok(())
}

#[derive(Parser)]
struct Args {
    #[clap(long)]
    db: PathBuf,

    #[clap(subcommand)]
    cmd: Cmd,
}

impl Args {
    fn run(&self) -> Result<(), BrokerError> {
        match &self.cmd {
            Cmd::Counter(x) => x.run(self)?,
            Cmd::Event(x) => x.run(self)?,
        }
        Ok(())
    }

    fn open_db(&self) -> Result<Db, BrokerError> {
        Db::new(&self.db).map_err(|e| BrokerError::Sqlite(self.db.clone(), e))
    }
}

#[derive(Parser)]
enum Cmd {
    Counter(Counter),
    Event(Event),
}

#[derive(Parser)]
struct Counter {
    #[clap(subcommand)]
    cmd: CounterCmd,
}

impl Counter {
    fn run(&self, args: &Args) -> Result<(), BrokerError> {
        match &self.cmd {
            CounterCmd::Show(x) => x.run(args)?,
            CounterCmd::Count(x) => x.run(args)?,
        }
        Ok(())
    }
}

#[derive(Parser)]
enum CounterCmd {
    Show(ShowCounter),
    Count(CountCounter),
}

#[derive(Parser)]
struct ShowCounter {}

impl ShowCounter {
    fn run(&self, args: &Args) -> Result<(), BrokerError> {
        let db = args.open_db()?;
        let counter = db
            .get_counter()
            .map_err(|e| BrokerError::Sqlite(args.db.clone(), e))?;
        println!("{}", counter.unwrap_or(0));
        Ok(())
    }
}

#[derive(Parser)]
struct CountCounter {
    #[clap(long)]
    goal: i64,
}

impl CountCounter {
    fn run(&self, args: &Args) -> Result<(), BrokerError> {
        let db = args.open_db()?;
        Self::inc(&db, self.goal).map_err(|e| BrokerError::Sqlite(args.db.clone(), e))?;

        Ok(())
    }

    fn inc(db: &Db, goal: i64) -> Result<(), sqlite::Error> {
        let mut prev: i64 = -1;
        loop {
            db.begin()?;
            println!("BEGIN");

            let current = db.get_counter()?;
            println!("  current as read={current:?}");
            let current = current.unwrap_or(0);
            println!("  current: {current}; prev={prev}");
            if current < prev {
                panic!("current < prev");
            }
            if current >= goal {
                println!("GOAL");
                db.rollback()?;
                println!("ROLLBACK");
                break;
            }

            let new = current + 1;
            if (new == 1 && db.create_counter(new).is_err()) || db.update_counter(new).is_err() {
                db.rollback()?;
                println!("ROLLBACK");
            } else {
                println!("  increment to {new}");
                db.commit()?;
                println!("COMMIT");
                prev = new;
            }
        }

        Ok(())
    }
}

#[derive(Parser)]
struct Event {
    #[clap(subcommand)]
    cmd: EventCmd,
}

impl Event {
    fn run(&self, args: &Args) -> Result<(), BrokerError> {
        match &self.cmd {
            EventCmd::Add(x) => x.run(args)?,
            EventCmd::List(x) => x.run(args)?,
            EventCmd::Show(x) => x.run(args)?,
            EventCmd::Remove(x) => x.run(args)?,
        }
        Ok(())
    }
}

#[derive(Parser)]
enum EventCmd {
    List(ListEvents),
    Add(AddEvent),
    Show(ShowEvent),
    Remove(RemoveEvent),
}

#[derive(Parser)]
struct ListEvents {}

impl ListEvents {
    fn run(&self, args: &Args) -> Result<(), BrokerError> {
        let db = args.open_db()?;
        for e in db
            .queued_events()
            .map_err(|e| BrokerError::Sqlite(args.db.clone(), e))?
        {
            println!("{e}");
        }
        Ok(())
    }
}

#[derive(Parser)]
struct AddEvent {
    #[clap(long)]
    repo: RepoId,

    #[clap(long, alias = "ref")]
    name: String,

    #[clap(long)]
    commit: Oid,

    #[clap(long)]
    base: Option<Oid>,

    #[clap(long)]
    output: Option<PathBuf>,
}

impl AddEvent {
    fn run(&self, args: &Args) -> Result<(), BrokerError> {
        let name = RefString::try_from(self.name.as_str()).expect("parse name as RefString");
        let event = BrokerEvent::new(&self.repo, &name, &self.commit, self.base);

        let db = args.open_db()?;
        let id = db
            .push_queued_event(event)
            .map_err(|e| BrokerError::Sqlite(args.db.clone(), e))?;
        println!("{id}");

        if let Some(filename) = &self.output {
            write(filename, id.to_string().as_bytes()).expect("write id file");
        }
        Ok(())
    }
}

#[derive(Parser)]
struct ShowEvent {
    #[clap(long, required_unless_present = "file")]
    id: Option<QueueId>,

    #[clap(long)]
    file: Option<PathBuf>,
}

impl ShowEvent {
    fn run(&self, args: &Args) -> Result<(), BrokerError> {
        let db = args.open_db()?;

        let id = if let Some(id) = &self.id {
            id.clone()
        } else {
            assert!(self.file.is_some());
            let file = self.file.as_ref().unwrap();
            let id = read(file).expect("read id file");
            let id = String::from_utf8_lossy(&id).to_string();
            QueueId::from(&id)
        };

        if let Some(event) = db
            .get_queued_event(&id)
            .map_err(|e| BrokerError::Sqlite(args.db.clone(), e))?
        {
            println!("{event:#?}");
        }
        Ok(())
    }
}

#[derive(Parser)]
struct RemoveEvent {
    #[clap(long, required_unless_present = "file")]
    id: Option<QueueId>,

    #[clap(long)]
    file: Option<PathBuf>,
}

impl RemoveEvent {
    fn run(&self, args: &Args) -> Result<(), BrokerError> {
        let db = args.open_db()?;

        let id = if let Some(id) = &self.id {
            id.clone()
        } else {
            assert!(self.file.is_some());
            let file = self.file.as_ref().unwrap();
            let id = read(file).expect("read id file");
            let id = String::from_utf8_lossy(&id).to_string();
            QueueId::from(&id)
        };

        db.remove_queued_event(&id)
            .map_err(|e| BrokerError::Sqlite(args.db.clone(), e))?;
        Ok(())
    }
}
