//! Database abstraction for the Radicle CI broker.
//!
//! This module is a wrapper around the [SQLite](https://sqlite.org/)
//! database. It is meant to suffice for the Radicle CI broker, and
//! does not try to be a more generic wrapper.
//!
//! The database stores the following kinds of data:
//!
//! - "counter": This is used for testing the implementation of
//!   concurrent access to the database. It is not useful for anything
//!   else.

use std::{
    fmt,
    path::Path,
    time::{Duration, Instant},
};

use sqlite::{Connection, State, Statement};
use uuid::Uuid;

use crate::event::BrokerEvent;

const MAX_WAIT: u64 = 1000; // how long to retry when SQL fails for busy database

/// The CI broker database. It stores the data that needs to be
/// persistent, even if the process terminates.
pub struct Db {
    conn: Connection,
}

impl Db {
    /// Open or create a database. It is created if it doesn't exist.
    /// If it is created, tables are created.
    pub fn new<P: AsRef<Path>>(filename: P) -> Result<Self, sqlite::Error> {
        let filename = filename.as_ref();
        let db = Db {
            conn: sqlite::open(filename)?,
        };

        {
            // let mut wal_stmt = Self::prepare(&db.conn, "PRAGMA journal_mode = WAL;")?;
            // Self::execute_valueless(&mut wal_stmt)?;

            db.begin()?;
            db.create_tables()?;
            db.commit()?;
        }

        Ok(db)
    }

    fn create_tables(&self) -> Result<(), sqlite::Error> {
        const TABLES: &[&str] = &[
            "CREATE TABLE IF NOT EXISTS counter_test (counter INT)",
            "CREATE TABLE IF NOT EXISTS event_queue (id TEXT PRIMARY KEY, timestamp TEXT, event TEXT, status TEXT)",
        ];

        for table in TABLES.iter() {
            let mut stmt = Self::prepare(&self.conn, table)?;
            Self::execute_valueless(&mut stmt)?;
        }

        Ok(())
    }

    /// Start a transaction.
    pub fn begin(&self) -> Result<(), sqlite::Error> {
        let mut stmt = Self::prepare(&self.conn, "BEGIN TRANSACTION")?;
        Self::execute_valueless(&mut stmt)
    }

    /// Commit a transaction.
    pub fn commit(&self) -> Result<(), sqlite::Error> {
        let mut stmt = Self::prepare(&self.conn, "COMMIT")?;
        Self::execute_valueless(&mut stmt)
    }

    /// Roll back a transaction.
    pub fn rollback(&self) -> Result<(), sqlite::Error> {
        let mut stmt = Self::prepare(&self.conn, "ROLLBACK")?;
        Self::execute_valueless(&mut stmt)
    }

    // Prepare a statement for execution.
    fn prepare<'a>(c: &'a Connection, sql: &str) -> Result<Statement<'a>, sqlite::Error> {
        let started = Instant::now();
        let max_wait = Duration::from_millis(MAX_WAIT);
        loop {
            match c.prepare(sql) {
                Ok(stmt) => {
                    return Ok(stmt);
                }

                // An error about the database being busy (locked) is
                // probably due to another process modifying the same
                // database. We ignore it as the problem should go away in
                // a very short while.
                Err(e) if e.code.unwrap() == sqlite3_sys::SQLITE_BUSY as isize => {
                    if started.elapsed() < max_wait {
                        continue;
                    }
                    return Err(e);
                }

                Err(e) => {
                    Err(e)?;
                }
            }
        }
    }

    // Execute a statement that doesn't return any rows with values.
    // This means basically any statement except S
    fn execute_valueless(stmt: &mut Statement) -> Result<(), sqlite::Error> {
        stmt.reset()?;

        let started = Instant::now();
        let max_wait = Duration::from_millis(MAX_WAIT);
        loop {
            match stmt.next() {
                Ok(_) => {
                    break;
                }

                // An error about the database being busy (locked) is
                // probably due to another process modifying the same
                // database. We ignore it as the problem should go away in
                // a very short while.
                Err(e) if e.code.unwrap() == sqlite3_sys::SQLITE_BUSY as isize => {
                    if started.elapsed() < max_wait {
                        continue;
                    }
                    return Err(e);
                }

                Err(e) => {
                    return Err(e);
                }
            }
        }

        Ok(())
    }

    /// Create the counter with an initial value. Only use this if
    /// there isn't a counter row already.
    pub fn create_counter(&self, counter: i64) -> Result<(), sqlite::Error> {
        let mut insert = self
            .conn
            .prepare("INSERT INTO counter_test (counter) VALUES (:1)")
            .unwrap();
        insert.bind((1, counter)).unwrap();
        match insert.next() {
            Ok(_) => (),
            Err(e) => return Err(e),
        }
        Ok(())
    }

    /// Update the counter to have a new value.
    pub fn update_counter(&self, counter: i64) -> Result<(), sqlite::Error> {
        let mut insert = self
            .conn
            .prepare("UPDATE counter_test SET counter = :1")
            .unwrap();
        insert.bind((1, counter)).unwrap();
        match insert.next() {
            Ok(_) => (),
            Err(e) => return Err(e),
        }
        Ok(())
    }

    /// Return the current value of the counter, if any.
    pub fn get_counter(&self) -> Result<Option<i64>, sqlite::Error> {
        let mut select = Self::prepare(&self.conn, "SELECT counter FROM counter_test")?;
        let mut counter = None;

        let started = Instant::now();
        let max_wait = Duration::from_millis(MAX_WAIT);

        loop {
            match select.next() {
                Ok(State::Row) => {
                    counter = Some(select.read("counter").unwrap());
                }
                Ok(State::Done) => {
                    break;
                }
                Err(e) if e.code.unwrap() == sqlite3_sys::SQLITE_BUSY as isize => {
                    if started.elapsed() < max_wait {
                        continue;
                    }
                    return Err(e);
                }
                Err(e) => {
                    return Err(e);
                }
            }
        }

        Ok(counter)
    }

    /// Return list of broker events currently in the queue.
    pub fn queued_events(&self) -> Result<Vec<QueueId>, sqlite::Error> {
        let mut select = Self::prepare(&self.conn, "SELECT id FROM event_queue")?;

        let started = Instant::now();
        let max_wait = Duration::from_millis(MAX_WAIT);

        let mut ids = vec![];

        loop {
            match select.next() {
                Ok(State::Row) => {
                    let id: String = select.read("id")?;
                    ids.push(QueueId::from(&id));
                }
                Ok(State::Done) => {
                    break;
                }
                Err(e) if e.code.unwrap() == sqlite3_sys::SQLITE_BUSY as isize => {
                    if started.elapsed() < max_wait {
                        continue;
                    }
                    return Err(e);
                }
                Err(e) => {
                    return Err(e);
                }
            }
        }

        Ok(ids)
    }

    /// Return a specific event, given is id, if one exists.
    pub fn get_queued_event(&self, id: &QueueId) -> Result<Option<BrokerEvent>, sqlite::Error> {
        let mut select = Self::prepare(&self.conn, "SELECT event FROM event_queue WHERE id = :id")?;
        select.bind((":id", id.as_str()))?;

        let started = Instant::now();
        let max_wait = Duration::from_millis(MAX_WAIT);

        let mut event = None;

        loop {
            match select.next() {
                Ok(State::Row) => {
                    let json: String = select.read("event")?;
                    event = Some(serde_json::from_str(&json).expect("parse BrokerEvent from JSON"));
                }
                Ok(State::Done) => {
                    break;
                }
                Err(e) if e.code.unwrap() == sqlite3_sys::SQLITE_BUSY as isize => {
                    if started.elapsed() < max_wait {
                        continue;
                    }
                    return Err(e);
                }
                Err(e) => {
                    return Err(e);
                }
            }
        }

        Ok(event)
    }

    /// Add a new event to the event queue, returning its id.
    pub fn push_queued_event(&self, event: BrokerEvent) -> Result<QueueId, sqlite::Error> {
        let json = serde_json::to_string(&event).expect("serialize BrokerEvent to JSON");

        let id = QueueId::default();
        let ts = "FIXME";
        let status = "FIXME";

        let mut insert = self
            .conn
            .prepare(
                "INSERT INTO event_queue (id, timestamp, event, status) VALUES (:id, :ts, :e, :s)",
            )
            .unwrap();
        insert.bind((":id", id.as_str())).unwrap();
        insert.bind((":ts", ts)).unwrap();
        insert.bind((":e", json.as_str())).unwrap();
        insert.bind((":s", status)).unwrap();
        match insert.next() {
            Ok(_) => (),
            Err(e) => return Err(e),
        }

        Ok(id)
    }

    /// Remove event from queue, given its id. It's OK if the event is
    /// not in the queue, that is just silently ignored.
    pub fn remove_queued_event(&self, id: &QueueId) -> Result<(), sqlite::Error> {
        let mut insert = self
            .conn
            .prepare("DELETE FROM event_queue WHERE id = :id")
            .unwrap();
        insert.bind((":id", id.as_str())).unwrap();

        match insert.next() {
            Ok(_) => (),
            Err(e) => return Err(e),
        }

        Ok(())
    }
}

/// An identifier for an event in the event queue in the database.
#[derive(Clone)]
pub struct QueueId {
    id: String,
}

impl fmt::Display for QueueId {
    fn fmt(&self, f: &mut fmt::Formatter) -> Result<(), fmt::Error> {
        write!(f, "{}", self.id)
    }
}

impl Default for QueueId {
    fn default() -> Self {
        Self {
            id: Uuid::new_v4().to_string(),
        }
    }
}

impl From<&str> for QueueId {
    fn from(id: &str) -> Self {
        Self { id: id.into() }
    }
}

impl From<&String> for QueueId {
    fn from(id: &String) -> Self {
        Self { id: id.into() }
    }
}

impl QueueId {
    fn as_str(&self) -> &str {
        self.id.as_str()
    }
}
